import os
import sys
import json
import struct

def dump(data, name) :
    with open(name, 'wb') as f :
        f.write(data)

# def readBits(data, bitOffset, bitLen) :
#     result = 0
#     readBitsCount = 0

#     while readBitsCount < bitLen :
#         byteOffset = int(bitOffset / 8)

#         availableBits = 8 - (bitOffset % 8)
#         bitsToRead = min(availableBits, bitLen - readBitsCount)
#         bitShift = availableBits - bitsToRead
#         print(bitShift)

#         mask = (1 << bitsToRead) - 1
#         result = (result << readBitsCount) | ((data[byteOffset] >> bitShift) & mask)
#         readBitsCount += bitsToRead
#         bitOffset     += bitsToRead

#     return result

def readBits(data, bitOffset, bitLen) :
    result = 0
    readBitsCount = 0

    while readBitsCount < bitLen :
        byteOffset = int(bitOffset / 8)

        availableBits = 8 - (bitOffset % 8)
        bitsToRead = min(availableBits, bitLen - readBitsCount)
        
        binValueStr = '{:08b}'.format(data[byteOffset])
        extractedBits = int(binValueStr[bitOffset%8:(bitOffset%8)+bitsToRead], 2)

        result = (result << bitsToRead) | extractedBits
        readBitsCount += bitsToRead
        bitOffset     += bitsToRead

    return result
    
def updateHuffmanTree(tree, item) :
    node = tree
    for i in range(item['code_length'] - 1, -1, -1) :
        bit = (item['code'] >> i) & 1
        if bit not in node :
            node[bit] = {}

        if i == 0 :
            node[bit] = item['symbol']
        else :
            node = node[bit]

    
    return tree

def parseStream1(data, literalBytesCount) :

    bitOffset = 0

    # first byte contains the index of the last table value
    huffTableSize = readBits(data, bitOffset, 8) + 1
    bitOffset += 8

    huffmanTree = {}
    huffmanTable = []
    code = 0
    currentCodeLength = -1
    for i in range(huffTableSize) :
        item = {}

        item['symbol'] = readBits(data, bitOffset, 8)
        bitOffset += 8

        item['code_length'] = readBits(data, bitOffset, 4) + 1
        bitOffset += 4

        if item['code_length'] == currentCodeLength :
            code += 1
        else :
            if currentCodeLength != -1 :
                code = (code + 1) << (item['code_length'] - currentCodeLength)
            currentCodeLength = item['code_length']

        item['code'] = code

        huffmanTable.append(item)
        huffmanTree = updateHuffmanTree(huffmanTree, item)

    # print(json.dumps(huffmanTable, indent = 4))
    # print(json.dumps(huffmanTree, indent = 4))

    # align to byte boundary after huffman table    
    if bitOffset % 8 != 0 :
        bitOffset += 8 - (bitOffset % 8)

    # read huffman encoded literal bytes
    literalBytes = b''

    # while bitOffset < (len(data) * 8) :
    for i in range(literalBytesCount) :
        node = huffmanTree[readBits(data, bitOffset, 1)]
        bitOffset += 1
        while type(node) == dict :
            if bitOffset >= (len(data) * 8) : break
            node = node[readBits(data, bitOffset, 1)]
            bitOffset += 1

        if type(node) == int :
            literalBytes += bytes([node])

    # print(literalBytes)

    return literalBytes

def parseTable(data, repeatedStringsCount, outTable) :
    data = data[0:]
    bitOffset = 0

    # first byte contains the index of the last table value
    huffTableSize = readBits(data, bitOffset, 5) + 1
    print('huffTableSize', huffTableSize)
    bitOffset += 5

    huffmanTree = {}
    huffmanTable = []
    code = 0
    currentCodeLength = -1
    for i in range(huffTableSize) :
        item = {}

        item['symbol'] = readBits(data, bitOffset, 5)
        bitOffset += 5

        item['code_length'] = readBits(data, bitOffset, 4) + 1
        bitOffset += 4

        if item['code_length'] == currentCodeLength :
            code += 1
        else :
            if currentCodeLength != -1 :
                code = (code + 1) << (item['code_length'] - currentCodeLength)
            currentCodeLength = item['code_length']

        item['code'] = code

        huffmanTable.append(item)
        huffmanTree = updateHuffmanTree(huffmanTree, item)

    print(json.dumps(huffmanTable, indent = 4))
    print(json.dumps(huffmanTree, indent = 4))

    # align to byte boundary after huffman table    
    if bitOffset % 8 != 0 :
        bitOffset += 8 - (bitOffset % 8)

    # read huffman encoded data
    for i in range(repeatedStringsCount) :
        print(data, bitOffset)
        node = huffmanTree[readBits(data, bitOffset, 1)]
        bitOffset += 1
        while type(node) == dict :
            if bitOffset >= (len(data) * 8) : break
            node = node[readBits(data, bitOffset, 1)]
            bitOffset += 1

        if type(node) == int :

            if node >= 16 :
                msbIndex = node - 12
                extraBits = readBits(data, bitOffset, msbIndex)
                bitOffset += msbIndex
                value = extraBits + (1 << msbIndex)
                outTable.append(value)
            else :
                outTable.append(node)

    # align to byte boundary huffman encoded data   
    if bitOffset % 8 != 0 :
        bitOffset += 8 - (bitOffset % 8)

    return bitOffset

def parseStream2(data, repeatedStringsCount, literalBytes) :

    offset = 0

    countOfPrecedingLiteralBytes = []
    consumedBits = parseTable(data[offset:], repeatedStringsCount, countOfPrecedingLiteralBytes)
    offset += int(consumedBits / 8)
    print('countOfPrecedingLiteralBytes', countOfPrecedingLiteralBytes)

    distances = []
    consumedBits = parseTable(data[offset:], repeatedStringsCount, distances)
    offset += int(consumedBits / 8)
    print('distances', distances)

    lengths = []
    consumedBits = parseTable(data[offset:], repeatedStringsCount, lengths)
    offset += int(consumedBits / 8)
    print('lengths', lengths)

    assert offset == len(data)

    g_decompressed = b''
    decompressedOffset = 0
    literalBytesOffset = 0
    for i in range(repeatedStringsCount) :
        g_decompressed += literalBytes[literalBytesOffset:literalBytesOffset+countOfPrecedingLiteralBytes[i]]
        literalBytesOffset += countOfPrecedingLiteralBytes[i]
        decompressedOffset += countOfPrecedingLiteralBytes[i]

        for j in range(lengths[i]) :
            assert distances[i] <= len(g_decompressed)
            g_decompressed += bytes([g_decompressed[-distances[i]]])

    # print(g_decompressed)
    return g_decompressed

def handleCompressedChunk(data) :
    offset = 0

    # XX XX XX
    # Literal bytes count
    literalBytesCount = struct.unpack_from('<H', data, offset)[0] + (data[offset + 2] << 16)
    offset += 3

    # XX XX XX
    streamSize = struct.unpack_from('<H', data, offset)[0] + (data[offset + 2] << 16)
    offset += 3

    stream1 = data[offset : offset+streamSize]
    offset += streamSize
    # dump(stream1, 'stream1.bin')
    literalBytes = parseStream1(stream1, literalBytesCount)

    # XX XX XX
    repeatedStringsCount = struct.unpack_from('<H', data, offset)[0] + (data[offset + 2] << 16)
    offset += 3
    # print('repeatedStringsCount', hex(repeatedStringsCount))
    

    stream2 = data[offset:]
    try :
        decompressedChunk = parseStream2(stream2, repeatedStringsCount, literalBytes)
    except :
        decompressedChunk = literalBytes
    # dump(stream2, 'stream2.bin')

    return decompressedChunk

def main() :

    with open(sys.argv[1], 'rb') as f :
        data = f.read()

    fout = open(sys.argv[1] + '.inflate', 'wb')

    offset = 0
    while offset < len(data) :

        # XX XX XX
        # Compressed chunk length
        compressedChunkSize = struct.unpack_from('<H', data, offset)[0] + (data[offset + 2] << 16)
        offset += 3

        # Compressed chunk
        compressedChunk = data[offset : offset+compressedChunkSize]
        offset += compressedChunkSize

        decomp = handleCompressedChunk(compressedChunk)
        fout.write(decomp)
        for i in range(len(decomp), 0x10000) :
            fout.write(b'\x00')

        dump(decomp, 'dump/bytes-{}.bin'.format(offset))

    assert offset == len(data)

    fout.close()

    return

if __name__ == '__main__' :
    main()