## Step 4

[LZ77 and LZ78](https://en.wikipedia.org/wiki/LZ77_and_LZ78)

Since LZ77 encodes and decodes from a sliding window over previously seen characters, decompression must always start at the beginning of the input.

As of 2008, the most popular LZ77-based compression method is DEFLATE; it combines LZSS with Huffman coding

[An Explanation of the Deflate Algorithm](https://www.zlib.net/feldspar.html)

[RFC 1951 - DEFLATE Compressed Data Format Specification version](http://www.faqs.org/rfcs/rfc1951.html)

https://github.com/madler/infgen/


### VM

  - rsi : bytecode pointer

ida vm1.py
```asm
label_0000 *(ctx) = var_00                   ; ctx             var_0a_mem + 4 * 0x0c
label_0007 *(file_buf) = var_01              ; file_buf        var_0a_mem + 4 * 0x0b
label_000e *(file_size) = var_02             ; file_size       var_0a_mem + 4 * 0x0a
label_0015 *(p_out_current) = var_03         ; p_out_current   var_0a_mem + 4 * 0x09
label_001c var_00 = *(p_out_current)
label_0023 *(p_out_compressed) = var_00      ;                 var_0a_mem + 4 * 0x08
label_002a var_00 = 0x10000
label_0033 *(chunk_size) = var_00 ; 0x10000  ;                 var_0a_mem + 4 * 0x07
label_003a var_00 = 0x0
label_0043 *(file_size_processed) = var_00 ;                   var_0a_mem + 4 * 0x06
label_004a jmp label_0050

label_0050 var_00 = *(file_size_processed)
label_0057 var_01 = *(file_size)
label_005e var_00 = var_00 < var_01
label_0065 jz var_00 -> label_0180           ; jmp to end if all input is processed

label_006c jmp label_0072

label_0072 var_00 = *(chunk_size)
label_0079 var_01 = *(file_size)
label_0080 var_02 = *(file_size_processed)
label_0087 var_01 = var_01 - var_02 ; file_size_remain
label_008e var_0e -= 4 * 1
label_0093 var_00 = call 0x000028fa_min(var_00, var_01) ; min(chunk_size, file_size_remain)
label_009b var_0e = var_0e + 4 * 0x01
label_00a2 *(var_0a_mem + 4 * 0x05) = var_00
label_00a9 var_00 = *(p_out_current)
label_00b0 *(var_0a_mem + 4 * 0x04) = var_00
label_00b7 var_00 = *(p_out_current)
label_00be var_01 = 0x3
label_00c7 var_00 = var_00 if 0x01 == 0xff else var_00 + var_01
label_00ce *(p_out_current) = var_00 ; p_out_current + 3
label_00d5 var_00 = *(ctx)
label_00dc var_01 = *(file_buf)
label_00e3 var_02 = *(file_size_processed)
label_00ea var_03 = *(var_0a_mem + 4 * 0x05)
label_00f1 var_03 = var_02 if 0x03 == 0xff else var_02 + var_03
label_00f8 var_04 = *(p_out_current)
label_00ff var_0e -= 4 * 1
label_0104 var_00 = call 0x00002e9c(var_00, var_01, var_02, var_03, var_04) ; call(ctx, file_buf, file_size_processed, file_size_to_process, p_out_chunk)
label_010c var_0e = var_0e + 4 * 0x01
label_0113 *(compressed_chunk_size) = var_00 ; result = compressed_chunk_size var_0a_mem + 4 * 0x03
label_011a var_00 = *(var_0a_mem + 4 * 0x04)
label_0121 var_01 = *(compressed_chunk_size)
label_0128 var_0e -= 4 * 1
label_012d var_00 = call 0x00001a57_write_word(var_00, var_01) ; write_word(p_out_current_backup, compressed_chunk_size)
label_0135 var_0e = var_0e + 4 * 0x01
label_013c var_00 = *(compressed_chunk_size)
label_0143 var_01 = *(p_out_current)
label_014a var_00 = var_01 if 0x00 == 0xff else var_01 + var_00
label_0151 *(p_out_current) = var_00 ; p_out_current += compressed_chunk_size
label_0158 jmp label_015e

label_015e var_00 = *(chunk_size)
label_0165 var_01 = *(file_size_processed)
label_016c var_00 = var_01 if 0x00 == 0xff else var_01 + var_00
label_0173 *(file_size_processed) = var_00 ; chunk_size + file_size_processed
label_017a jmp label_0050

label_0180 var_00 = *(p_out_current)
label_0187 var_01 = *(p_out_compressed)
label_018e var_00 = var_00 - var_01
label_0195 eax = var_00 ; ret
```

  - input chunk de 0x10000
  - format compressé : <3-bytes compressed size chunk><compressed chunk>


python .\inflate.py .\home_backup.tar.zz

dans les chunks decompressés seulement avec le codage de huffman :

$ strings dump/bytes-* | grep SSTIC
SSTIC{0ded220ffb9d4215b090ebb509e7a1ef}


$ python .\inflate.py .\home_backup.tar.zz

encore des erreurs de décompression mais semble suffisant pour continuer

$ tar xvf home_backup.tar.zz.inflate
./FIORANELLI-RETRACTED.pdf
./notes.txt
./lobster_dog.tga
.bash_history

$ cat notes.txt 
J'ai lu le papier de FIORANELLI et effectivement nous avons
bien fait de le faire rétracter, un peu plus et il aurait été
pris au sérieux et aurait attiré l'attention du grand public...

De telles informations auraient pu réduire l'Organisation
à néant...

SSTIC{0ded220ffb9d4215b090ebb509e7a1ef}

$ cat .bash_history 
ls -la
whoami
id
cd /tmp
ls
mounter_client mount goodfs MGhtT34gHj5yFcszRYB4gf45DtymEi
cd /mnt/goodfs
ls
cd
cd
cd
exit
